/* $Id: et6000.c,v 1.10 1999/06/21 23:44:57 marcus Exp $
******************************************************************************

   KGI - Tseng Labs ET6000 accelerated graphics driver

   Copyright (C) 1997	 Jim Ursetto	[jim.ursetto@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code is based on the following:
   [1] XFree86 ET6000 accelerated support by Koen Gadeyne
   [2] Trio64 driver by Jon Taylor

******************************************************************************
*/

#include <kgi/maintainers.h>

#define DRIVER_NAME	"Tseng Labs ET6000 graphics driver"
#define	DRIVER_REV	"$Revision: 1.10 $"

/* Take this out to prevent acceleration. */
#define USE_ACCEL

/* KGI includes */
#include <kgi/module.h>
#include <kgi/kgi_commands.h>

#include "Tseng/accel.h"

int acl_bpp;

const char *sugg_default[KGIGT_LAST][5]= {
	{ "generic-text-16","generic-ramdac",NULL,NULL,NULL },
	{ "generic-text-32","generic-ramdac",NULL,NULL,NULL },
	{ "generic-linear-1","generic-ramdac",NULL,NULL,NULL },
	{ "generic-linear-4","generic-ramdac",NULL,NULL,NULL},
	{ "generic-linear-8","generic-ioctl","generic-ramdac",NULL,NULL},
	{ "generic-linear-15","generic-ioctl",NULL,NULL,NULL},
	{ "generic-linear-16","generic-ioctl",NULL,NULL,NULL},
	{ "generic-linear-24","generic-ioctl","generic-gamma",NULL,NULL},
	{ "generic-linear-32","generic-ioctl","generic-gamma",NULL,NULL}
};

struct chipset_private
{
        kgiu32 linear_mem_size;
};

#define GET_PRIVATE \
            struct chipset_private *priv = KGIM_PRIVATE(dpy, chipset)

#define acl_in8(reg) \
            mem_in8(accelRegs + reg)
#define acl_out8(val, reg) \
            mem_out8(val, accelRegs + reg)
#define acl_out16(val, reg) \
            mem_out16(val, accelRegs + reg)
#define acl_out32(val, reg) \
            mem_out32(val, accelRegs + reg)

/* XWindows raster ops */
int RopTable[] = {
    0x00, /* Xclear		0 */
    0x88, /* Xand		src AND dst */
    0x44, /* XandReverse	src AND NOT dst */
    0xcc, /* Xcopy		src */
    0x22, /* XandInverted	NOT src AND dst */
    0xaa, /* Xnoop		dst */
    0x66, /* Xxor		src XOR dst */
    0xee, /* Xor		src OR dst */
    0x11, /* Xnor		NOT src AND NOT dst */
    0x99, /* Xequiv		NOT src XOR dst */
    0x55, /* Xinvert		NOT dst */
    0xdd, /* XorReverse		src OR NOT dst */
    0x33, /* XcopyInverted	NOT src */
    0xbb, /* XorInverted	NOT src OR dst */
    0x77, /* Xnand		NOT src OR NOT dst */
    0xff  /* Xset		1 */
};

int PatternRopTable[] = {
    0x00, /* Xclear		0 */
    0xa0, /* Xand		pat AND dst */
    0x50, /* XandReverse	pat AND NOT dst */
    0xf0, /* Xcopy		pat */
    0x0a, /* XandInverted	NOT pat AND dst */
    0xaa, /* Xnoop		dst */
    0x5a, /* Xxor		pat XOR dst */
    0xfa, /* Xor		pat OR dst */
    0x05, /* Xnor		NOT pat AND NOT dst */
    0xa5, /* Xequiv		NOT pat XOR dst */
    0x55, /* Xinvert		NOT dst */
    0xf5, /* XorReverse		pat OR NOT dst */
    0x0f, /* XcopyInverted	NOT pat */
    0xaf, /* XorInverted	NOT pat OR dst */
    0x5f, /* Xnand		NOT pat OR NOT dst */
    0xff  /* Xset		1 */
};

/* Pingpong buffers. At the moment this probably doesn't matter since
 * we can't queue accelerator operations yet.  Basically just here for
 * possible future use, and it may even slow things down now.
 * kgiu32 types are offsets into memory; mem_vaddr types are actual
 * memory addresses.  We need to write directly into memory, but the
 * accelerator needs the offsets.  I may take out the "ptr" variables
 * since they're just calculated from lfb + non-ptr variables.
 */

static kgiu32 SrcPing;
static mem_vaddr ptrSrcPing;
static kgiu32 DestPing;
static mem_vaddr ptrDestPing;
static kgiu32 PatPing;
static mem_vaddr ptrPatPing;

static kgiu32 SrcPong;
static mem_vaddr ptrSrcPong;
static kgiu32 DestPong;
static mem_vaddr ptrDestPong;
static kgiu32 PatPong;
static mem_vaddr ptrPatPong;

/* These point to the current ping or pong buffer. */
/* MixMap not added yet. */
static kgiu32 SrcMap;
static kgiu32 DestMap;
static kgiu32 PatMap;
static mem_vaddr ptrSrcMap;
static mem_vaddr ptrDestMap;
static mem_vaddr ptrPatMap;

static mem_vaddr accelRegs;
static mem_vaddr linmem;
static int fbwidth;

static inline void doPingPong(void)
{
    if (SrcMap == SrcPong) {
        SrcMap = SrcPing;
		ptrSrcMap = ptrSrcPing;
		DestMap = DestPing;
		ptrDestMap = ptrDestPing;
		PatMap = PatPing;
		ptrPatMap = ptrPatPing;
    } else {
    	SrcMap = SrcPong;
		ptrSrcMap = ptrSrcPong;
		DestMap = DestPong;
		ptrDestMap = ptrDestPong;
		PatMap = PatPong;
		ptrPatMap = ptrPatPong;
    }
    
}

static inline int acl_wait(int);

static void terminateAccel(int maysched)
{
    if (acl_in8(ACL_ACCELERATOR_STATUS) & (ACL36_BUSY | ACL36_XYUNSTABLE)) {
		acl_out8(0x00, ACL_SUSPEND_TERMINATE);
		acl_out8(ACL30_SUSPEND, ACL_SUSPEND_TERMINATE);
		acl_wait(maysched);
		acl_out8(0x00, ACL_SUSPEND_TERMINATE);
		acl_out8(ACL30_TERMINATE, ACL_SUSPEND_TERMINATE);
		acl_wait(maysched);
		acl_out8(0x00, ACL_SUSPEND_TERMINATE);
    }
	DEBUG2("graphics: Accelerator terminated.\n");
}

#define MAX_ACL_WAIT 500000    /* like XFree86 */

static inline int acl_wait(int maysched)
{
	int x = MAX_ACL_WAIT;
	while ((acl_in8(ACL_ACCELERATOR_STATUS) & ACL36_BUSY) && --x) {
		if (maysched) schedule();
	}
    
	if (!x)
		DEBUG1("graphics: accelerator timeout\n");
	return (!x);
}

static inline int replicate_color(int color)
{
    switch (acl_bpp) {
		case 1:
			color &= 0xFF;
			color = (color <<  8) | color;
			color = (color << 16) | color;
			break;
		case 2:
			color &= 0xFFFF;
			color = (color << 16) | color;
			break;
    }
    return color;
}

static inline void set_fgcolor(int color)
{
	acl_out32(SrcMap, ACL_SOURCE_ADDRESS);
	acl_out16(0x03, ACL_SOURCE_Y_OFFSET);  /* 4 bytes between lines */
	color = replicate_color(color);
	acl_out8(0x02, ACL_SOURCE_WRAP);  /* wrap every 4 bytes horiz, 1 line vert */
	mem_out32(color, ptrSrcMap);   /* put color in the source map */
}


/* Optimize this math */
static int et6000_drawbox(struct kgi_graphic *gr, void *iobuf, int maysched)
{	
	struct kgi_box *boxp = (struct kgi_box *)iobuf;

	if (gr->dev.mode.graphtype == KGIGT_4BIT || 
		gr->dev.mode.graphtype == KGIGT_TEXT16)
	{
		return -E(ACCEL,NOSUP_MODE_MMAP);
	}
	if (boxp->w && boxp->h) {	/* 0 width is not OK! */

	    doPingPong();

	    /* This can go in a "setup" function later. */
	    acl_out8(0xCC, ACL_FG_RASTER_OP); /* copy src */
	    set_fgcolor(gr->gc->fg_color);  /* GC passing finally implemented! */
		acl_out8(0x33, ACL_MIX_CONTROL); /* mix = fg, mask = 1 */

		/* Do the draw. */

	    acl_out8(0x00,ACL_XY_DIRECTION); /* low to high, bitblt */
	    /* Can optimize bpp mult later */
	    acl_out32((((boxp->h - 1) << 16) | (boxp->w*acl_bpp - 1)),ACL_X_COUNT);
	    acl_out32(boxp->x*acl_bpp + boxp->y*fbwidth, ACL_DEST_ADDRESS);
	    acl_wait(1);      /* temporary fix 
						   * to prevent subsequent lfb access
						   * from being clobbered by waiting operation
						   * but also prevents queueing */
	}
    return 0;
}

static int et6000_copybox(struct kgi_graphic *gr,void *iobuf)
{	
    struct kgi_cbox *boxp = (struct kgi_cbox *)iobuf;
    int dir = 0;
    int x, x2;

	if (gr->dev.mode.graphtype == KGIGT_4BIT || 
	    gr->dev.mode.graphtype == KGIGT_TEXT16)
		return -E(ACCEL,NOSUP_MODE_MMAP);

	if (boxp->w && boxp->h) {		/* 0 width is not OK! */

		boxp->w *= acl_bpp;  /* optimize! */
		x = boxp->x * acl_bpp; /* waste of cycles ! */
		x2 = boxp->x2 * acl_bpp;
	    
		if (boxp->x <= boxp->x2)  { /* they're going to overlap */
			dir |= 0x01;         /* sneak up from behind */
			x += boxp->w - 1;
			x2 += boxp->w - 1;
		}
		if (boxp->y <= boxp->y2) 
		{ 
			dir |= 0x02;
			boxp->y += boxp->h - 1;
			boxp->y2 += boxp->h - 1;
		}
	    
		/* Direction matters if the areas overlap. */
		acl_out8(dir, ACL_XY_DIRECTION);

		/* For transparency, when available. */
		/*      SET_BG_COLOR(transparency_color); 
				acl_out8(0x13, ACL_MIX_CONTROL); */
		/* mix = inv compare map */
	    
		acl_out8(0x33, ACL_MIX_CONTROL); /* mix = fg, mask = 1 */

		/* ROP - change when GC rop is available */
		acl_out8(0xCC, ACL_FG_RASTER_OP); /* copy src */	    
            
		acl_out8(0x77, ACL_SOURCE_WRAP); /* don't wrap, obviously */
            
		acl_out16(fbwidth - 1, ACL_SOURCE_Y_OFFSET); /* bytes between lines */
		acl_out32((((boxp->h - 1) << 16) | (boxp->w - 1)),ACL_X_COUNT);
		acl_out32(x + boxp->y*fbwidth, ACL_SOURCE_ADDRESS);
		acl_out32(x2 + boxp->y2*fbwidth, ACL_DEST_ADDRESS);
		acl_wait(1);
	}
	return(0);
}

/* KGI module interface */

int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	mem_vaddr scratchfb;
	kgiu32 scratchpad;
	mem_vaddr linear_mem;
	kgiu32 accel_regs;
	GET_PRIVATE;
    
	linear_mem = dpy->mode.fb;
	linmem = linear_mem;
    
	/* Only 4088 KB can be used in linear, accelerated mode. 
	 * We need 1 KB of video ram for a scratch buffer. 
	 * Accelerator registers are located at framebuffer + 0x3fff00. */
    
	/* Notify chipset driver that we're taking the scratchpad space. */
	priv->linear_mem_size -= 1 KB;

	scratchpad = priv->linear_mem_size;
	accel_regs  = 4096 KB - 256;
	DEBUG1("graphics: Linear framebuffer at 0x%p\n", linear_mem);
	DEBUG1("graphics: Scratchpad at lfb + 0x%08x\n", scratchpad);
	DEBUG1("graphics: Accelerator regs at lfb + 0x%08x\n", accel_regs);
    
    /* Set up the ping-pong buffers. */

	SrcPing  = scratchpad;
	SrcPong  = scratchpad + 8;
	DestPing = scratchpad + 16;
	DestPong = scratchpad + 24;
	PatPing  = scratchpad + 32;
	PatPong  = scratchpad + 40;

	scratchfb = linear_mem + scratchpad;    
    
	ptrSrcPing  = scratchfb;
	ptrSrcPong  = scratchfb + 8;
	ptrDestPing = scratchfb + 16;
	ptrDestPong = scratchfb + 24;
	ptrPatPing  = scratchfb + 32;
	ptrPatPong  = scratchfb + 40;

	accelRegs = linear_mem + accel_regs;
    
	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE1("ET6000 Graphics Driver removed");
}

int kgim_accel_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd *cmd)
{
	/* This is where I'd do the accelerator bandwidth check. */
	return EOK;
}

void kgim_accel_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	/* We'll set up the accelerator now. */
	switch (mode->request.graphtype) {
		case KGIGT_8BIT:
			acl_bpp = 1; break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
			acl_bpp = 2; break;
		case KGIGT_24BIT:
			acl_bpp = 3; break;
		case KGIGT_32BIT:
			acl_bpp = 4; break;
		default:
			return;
	}
	/* Set up initial pingpong buffers. */
	SrcMap = SrcPing;
	ptrSrcMap = ptrSrcPing;
	DestMap = DestPing;
	ptrDestMap = ptrDestPing;
	PatMap = PatPing;
	ptrPatMap = ptrPatPing;
	
	/* Kill the accelerator. */
    terminateAccel(0);
	
	/* The next two lines do nothing according to VGADOC. */
	acl_out8(0x0e, ACL_INTERRUPT_STATUS);  /* clear interrupts */
	acl_out8(0x04, ACL_INTERRUPT_MASK);    /* disable interrupts, but enable deadlock exit */
	acl_out8(0x00, ACL_INTERRUPT_STATUS);
	acl_out8(0x00, ACL_ACCELERATOR_STATUS);
	
	acl_out8(0x00, ACL_STEP_INHIBIT); /* Disable stepping inhibit. */
	acl_out8(0x00, ACL_CONFIG_6000);      /* Best performance. */
	acl_out8(ACL37_DOWN_WHEN_IDLE, ACL_POWER_CONTROL);
	acl_out8(0x00, ACL_XFER_DISABLE);     /* Enable all transfers. */

	acl_out16(0x00, ACL_X_POSITION);
	acl_out16(0x00, ACL_Y_POSITION);
	acl_out8((acl_bpp - 1) << 4, ACL_PIXEL_DEPTH);  /* Not sure about this one! */
	/* writing destination address will start ACL */
	acl_out8(ACL31_START_ENABLE, ACL_OPERATION_STATE);

	acl_out16(mode->request.virt.x * acl_bpp - 1, ACL_DEST_Y_OFFSET);
	acl_out8(0x00, ACL_XY_DIRECTION);

	/* We're not using the MMU apertures, but I'll set it anyway since XF86 does */
	acl_out8(0x74, MMU_CONTROL);
	fbwidth = mode->request.virt.x * acl_bpp;
}

int kgim_accel_command (struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{ 
	kgi_suggest *sugg;
	const char *name;

	switch (cmd) {

		case ACCEL_GETSUGGEST:
			sugg = (kgi_suggest *)(gr->io_buf);
			gr->io_res = gr->io_buf;
			if (sugg->id < 0 || sugg->id > 4) {
				return -E(ACCEL, BOGUS);
			}

	        name = sugg_default[gr->dev.mode.graphtype][sugg->id];
			if (!name) {
				return -E(ACCEL,BOGUS);
			}

			strcpy(sugg->name, name);
			sugg->args[0] = 0;
			sugg->id++;	/* Next is RAMDAC */

			/* LOOKHERE--may be a debugging remnant */
			if (sugg->id >= 4) {
				sugg->id = 0;
			}
	        if (sugg_default[gr->dev.mode.graphtype][sugg->id] == NULL) {
				sugg->id = 0;
			}
			return EOK;

#ifdef USE_ACCEL	    
		case ACCEL_DRAWBOX:
			return(et6000_drawbox(gr, gr->io_buf, 1));

		case ACCEL_COPYBOX:
			return(et6000_copybox(gr, gr->io_buf));

		case ACCEL_FILLSCREEN:
			return -E(ACCEL,NOSUP_ALWAYS_LOWER); 

#endif	  
		default:
		    break;
	}
	return -E(ACCEL, NOSUP_ALWAYS_MMAP);
}

#undef acl_in8
#undef acl_out8
#undef acl_out16
#undef acl_out32
